# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT license.
import re

from textworld.envs.glulx.git_glulx import GitGlulxEnv
from textworld.envs.zmachine.jericho import JerichoEnv
from textworld.envs.tw import TextWorldEnv
from textworld.envs.wrappers.tw_inform7 import TWInform7
from textworld.envs.pddl import PddlEnv


def _guess_backend(path):
    # Guess the backend from the extension.
    if path.endswith(".ulx"):
        return GitGlulxEnv
    elif re.search(r"\.z[1-8]", path):
        return JerichoEnv
    elif path.endswith(".json"):
        return TextWorldEnv
    elif path.endswith(".tw-pddl"):
        return PddlEnv

    msg = "Unsupported game format: {}".format(path)
    raise ValueError(msg)



import json
from copy import deepcopy
from collections import defaultdict

from tatsu.model import NodeWalker
from typing import List, Dict

from textworld.utils import check_flag
from textworld.logic import Rule
from textworld.envs.pddl.textgen.model import CSGModelBuilderSemantics
from textworld.envs.pddl.textgen.parser import CSGParser


class Symbol:
    def __init__(self, symbol: str, context: Dict = {}):
        self.symbol = symbol
        self.context = context

    def __str__(self):
        return str(self.symbol)

    def copy(self, context):
        copy = deepcopy(self)
        copy.context = copy_context(context)
        return copy


class TerminalSymbol(Symbol):

    def __repr__(self):
        return "TerminalSymbol({})".format(repr(self.symbol))


class NonterminalSymbol(Symbol):

    def __repr__(self):
        return "NonterminalSymbol('#{}#')".format(self.symbol)


def copy_context(context):
    return {
        "state": context["state"],
        "facts": context["facts"],
        "variables": deepcopy(context["variables"]),
        "mapping": deepcopy(context["mapping"]),
        "entity_infos": context["entity_infos"],
    }


def join(sep, iterable):
    """ Join an iterable with a separator similar to `str.join`. """
    iterable = iter(iterable)
    yield next(iterable)

    for e in iterable:
        yield sep
        yield e


def display_list(l, context):
    if len(l) == 0:
        return [NonterminalSymbol("list_empty", context)]

    if len(l) == 1:
        return [l[0]]

    list_separator = NonterminalSymbol("list_separator", context)
    list_last_separator = NonterminalSymbol("list_last_separator", context)
    return list(join(list_separator, l[:-1])) + [list_last_separator] + [l[-1]]


def query(expression, context):
    rule = Rule.parse_conjunctive_query(expression)

    contexts = []
    for mapping in context["state"].all_assignments(rule, context["mapping"]):
        context_ = copy_context(context)
        new_variables = {ph.name: context_["entity_infos"][var.name] for ph, var in mapping.items()}
        context_["variables"].update(new_variables)
        context_["mapping"].update(mapping)
        contexts.append(context_)

    return contexts


class EvalSymbol(Symbol):
    def __init__(self, expression: str, context: Dict = {}):
        super().__init__(expression, context)
        self.expression = expression

    def __repr__(self):
        return "EvalSymbol('{{{}}}')".format(str(self.expression))

    def derive(self, context=None):
        context = context or self.context
        locals().update(context["variables"])
        value = eval(self.expression)
        return [TerminalSymbol(value)]


class ListSymbol(Symbol):
    def __init__(self, symbol: Symbol, context: Dict = {}):
        super().__init__(symbol, context)

    def __repr__(self):
        return "ListSymbol('[{!r}]')".format(self.symbol)

    def derive(self, context=None):
        context = context or self.context
        self.symbol.context = context
        derivation = self.symbol.derive()
        return display_list(derivation, context)


class ConditionalSymbol(Symbol):

    def __init__(self, expression: Symbol, given: str, context: Dict = {}):
        super().__init__(str(expression), context)
        self.expression = expression
        self.given = given

    def __repr__(self):
        return "ConditionalSymbol('{{{}|{}}}')".format(str(self.expression), str(self.given))

    def derive(self):
        if len(self.context) == 0:
            raise ValueError("Empty context")

        context = copy_context(self.context)

        contexts = [context]
        if self.given:
            contexts = query(self.given, context)

        res = [self.expression.copy(context) for context in contexts]
        return res


class ProductionRule:
    """ Production rule for a context-sensitive grammar. """

    def __init__(self, lhs: str, rhs: List[Symbol], weight=1, condition=None):
        """
        Arguments:
            rhs: symbol that will be transformed by this production rule.
            lhs: list of symbols generated by this production rule.
            weight: prevalence of this production.
        """
        self.lhs = lhs
        self.rhs = rhs
        self.weight = weight
        self.condition = condition

    def __repr__(self):
        text = "ProductionRule(lhs={!r}, rhs={!r}, weight={!r}, condition={!r})"
        return text.format(self.lhs, self.rhs, self.weight, self.condition)


class _Converter(NodeWalker):

    def walk_list(self, node):
        return [self.walk(child) for child in node]

    def walk_str(self, node):
        return node.replace("\\n", "\n")

    def walk_TerminalSymbol(self, node):
        return TerminalSymbol(node.literal)

    def walk_NonterminalSymbol(self, node):
        return NonterminalSymbol(node.symbol)

    def walk_ConditionalSymbol(self, node):
        return ConditionalSymbol(self.walk(node.expression), node.given)

    def walk_EvalSymbol(self, node):
        return EvalSymbol(node.statement)

    def walk_ListSymbol(self, node):
        return ListSymbol(self.walk(node.symbol))


_PARSER = CSGParser(semantics=CSGModelBuilderSemantics(), parseinfo=True)


def _parse_and_convert(*args, **kwargs):
    model = _PARSER.parse(*args, **kwargs)
    return _Converter().walk(model)


class CSGUnknownSymbolError(Exception):
    def __init__(self, symbol: Symbol):
        msg = "Can't find symbol '#{}#' in the set of production rules."
        # TODO: mention closest match?
        super().__init__(msg.format(symbol))


class ContextSensitiveGrammar:

    def __init__(self):
        self._rules = defaultdict(list)

    def update(self, grammar: "ContextSensitiveGrammar"):
        for k, v in grammar._rules.items():
            self._rules[k].extend(v)

    @classmethod
    def parse(cls, text: str):
        data = json.loads(text)

        grammar = cls()
        for name, rules in data.items():
            rules = [ProductionRule(lhs=name,
                                    rhs=_parse_and_convert(rule["rhs"], rule_name="symbols"),
                                    weight=rule.get("weight", 1),
                                    condition=rule.get("condition", ""))
                     for rule in rules]

            for rule in rules:
                grammar.add_rule(rule)

        return grammar

    def add_rule(self, rule: ProductionRule):
        self._rules[rule.lhs].append(rule)

    def replace(self, start: Symbol) -> List[Symbol]:
        rules = self._rules.get(str(start))
        if not rules:
            raise CSGUnknownSymbolError(start)

        def _applicable(rule):
            if not rule.condition:
                return True

            rule = Rule.parse_conjunctive_query(rule.condition)
            mappings = list(start.context["state"].all_assignments(rule, start.context["mapping"]))
            return len(mappings) > 0  # Found a match.

        rules = list(filter(_applicable, rules))

        # TODO: deal with multiple alternatives
        # TODO: deal with context
        symbols = deepcopy(rules[0].rhs)
        for symbol in symbols:
            symbol.context = copy_context(start.context)

        return symbols

    def derive(self, start: str, context={}) -> str:
        derivation = _parse_and_convert(start, rule_name="symbols", trace=check_flag("TW_CSG_TRACE"))
        derivation = derivation[::-1]  # Reverse to build a derivation stack.

        for symbol in derivation:
            symbol.context = copy_context(context)

        derived = []
        while len(derivation) > 0:
            if check_flag("TW_CSG_DEBUG"):
                print(derivation)

            symbol = derivation.pop()
            if isinstance(symbol, TerminalSymbol):
                derived.append(symbol)

            elif isinstance(symbol, NonterminalSymbol):
                derivation += self.replace(symbol)[::-1]  # Reverse to add on top of the derivation stack.

            elif isinstance(symbol, (ConditionalSymbol, EvalSymbol, ListSymbol)):
                derivation += symbol.derive()[::-1]  # Reverse to add on top of the derivation stack.

            else:
                raise NotImplementedError("Unknown symbol: {}".format(type(symbol)))

        return "".join(map(str, derived))



#!/usr/bin/env python

# CAVEAT UTILITOR
#
# This file was automatically generated by TatSu.
#
#    https://pypi.python.org/pypi/tatsu/
#
# Any changes you make to it will be overwritten the next time
# the file is generated.

from __future__ import annotations

import sys

from tatsu.buffering import Buffer
from tatsu.parsing import Parser
from tatsu.parsing import tatsumasu
from tatsu.parsing import leftrec, nomemo, isname # noqa
from tatsu.infos import ParserConfig
from tatsu.util import re, generic_main  # noqa


KEYWORDS = {}  # type: ignore


class CSGBuffer(Buffer):
    def __init__(self, text, /, config: ParserConfig = None, **settings):
        config = ParserConfig.new(
            config,
            owner=self,
            whitespace=re.compile('[\\t]+'),
            nameguard=None,
            comments_re=None,
            eol_comments_re='^(//.*|\\s*)\\n?',
            ignorecase=False,
            namechars='',
            parseinfo=False,
        )
        config = config.replace(**settings)
        super().__init__(text, config=config)


class CSGParser(Parser):
    def __init__(self, /, config: ParserConfig = None, **settings):
        config = ParserConfig.new(
            config,
            owner=self,
            whitespace=re.compile('[\\t]+'),
            nameguard=None,
            comments_re=None,
            eol_comments_re='^(//.*|\\s*)\\n?',
            ignorecase=False,
            namechars='',
            parseinfo=False,
            keywords=KEYWORDS,
            start='start',
        )
        config = config.replace(**settings)
        super().__init__(config=config)

    @tatsumasu()
    def _start_(self):  # noqa
        self._symbols_()

    @tatsumasu()
    def _tag_(self):  # noqa
        self._pattern('[\\w()/!<>\\-\\s,.]+')

    @tatsumasu()
    def _given_(self):  # noqa
        self._pattern('[^;|{}\\n\\[\\]#]+')

    @tatsumasu()
    def _statement_(self):  # noqa
        self._pattern('[^|\\[\\]{}\\n<>]+')

    @tatsumasu()
    def _Literal_(self):  # noqa
        self._pattern('[^;|"<>\\[\\]#{}]*')

    @tatsumasu('TerminalSymbol')
    def _terminalSymbol_(self):  # noqa
        with self._group():
            with self._choice():
                with self._option():
                    self._token('"')
                    self._Literal_()
                    self.name_last_node('literal')
                    self._token('"')

                    self._define(
                        ['literal'],
                        []
                    )
                with self._option():
                    self._cut()
                    self._Literal_()
                    self.name_last_node('literal')

                    self._define(
                        ['literal'],
                        []
                    )
                self._error(
                    'expecting one of: '
                    '\'"\' \'~\''
                )

    @tatsumasu('NonterminalSymbol')
    def _nonterminalSymbol_(self):  # noqa
        self._token('#')
        self._tag_()
        self.name_last_node('symbol')
        self._token('#')

        self._define(
            ['symbol'],
            []
        )

    @tatsumasu('EvalSymbol')
    def _evalSymbol_(self):  # noqa
        self._statement_()
        self.name_last_node('statement')

    @tatsumasu('ConditionalSymbol')
    def _conditionalSymbol_(self):  # noqa
        self._token('{')
        with self._group():
            with self._choice():
                with self._option():
                    self._nonterminalSymbol_()
                with self._option():
                    self._evalSymbol_()
                self._error(
                    'expecting one of: '
                    '<evalSymbol> <nonterminalSymbol>'
                )
        self.name_last_node('expression')
        with self._optional():
            self._pattern('\\s*\\|\\s*')
            self._given_()
            self.name_last_node('given')

            self._define(
                ['given'],
                []
            )
        self._token('}')

        self._define(
            ['expression', 'given'],
            []
        )

    @tatsumasu('ListSymbol')
    def _listSymbol_(self):  # noqa
        self._token('[')
        self._conditionalSymbol_()
        self.name_last_node('symbol')
        self._token(']')

        self._define(
            ['symbol'],
            []
        )

    @tatsumasu()
    def _Symbol_(self):  # noqa
        with self._choice():
            with self._option():
                self._listSymbol_()
            with self._option():
                self._conditionalSymbol_()
            with self._option():
                self._nonterminalSymbol_()
            with self._option():
                self._terminalSymbol_()
            self._error(
                'expecting one of: '
                '\'"\' \'#\' \'[\' \'{\' \'~\' <conditionalSymbol>'
                '<listSymbol> <nonterminalSymbol>'
                '<terminalSymbol>'
            )

    @tatsumasu()
    def _symbols_(self):  # noqa

        def block0():
            self._Symbol_()
        self._positive_closure(block0)


class CSGSemantics:
    def start(self, ast):  # noqa
        return ast

    def tag(self, ast):  # noqa
        return ast

    def given(self, ast):  # noqa
        return ast

    def statement(self, ast):  # noqa
        return ast

    def Literal(self, ast):  # noqa
        return ast

    def terminalSymbol(self, ast):  # noqa
        return ast

    def nonterminalSymbol(self, ast):  # noqa
        return ast

    def evalSymbol(self, ast):  # noqa
        return ast

    def conditionalSymbol(self, ast):  # noqa
        return ast

    def listSymbol(self, ast):  # noqa
        return ast

    def Symbol(self, ast):  # noqa
        return ast

    def symbols(self, ast):  # noqa
        return ast


def main(filename, **kwargs):
    if not filename or filename == '-':
        text = sys.stdin.read()
    else:
        with open(filename) as f:
            text = f.read()
    parser = CSGParser()
    return parser.parse(
        text,
        filename=filename,
        **kwargs
    )


if __name__ == '__main__':
    import json
    from tatsu.util import asjson

    ast = generic_main(main, CSGParser, name='CSG')
    data = asjson(ast)
    print(json.dumps(data, indent=2))



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT license.


# -*- coding: utf-8 -*-
import json

from typing import Mapping, Union, Optional

import textworld
from textworld.core import EnvInfos, GameState
from textworld.generator.game import EntityInfo
from textworld.logic import Proposition, Variable

try:
    import fast_downward
    from textworld.envs.pddl import logic as pddl_logic
    fast_downward_missing = False
except ImportError:
    fast_downward_missing = True


class PddlEnv(textworld.Environment):
    """
    Environment for playing games defined by a PDDL file.
    """

    def __init__(self, infos: Optional[EnvInfos] = None) -> None:
        """
        Arguments:
            infos: Information to be included in the game state. By
                       default, only the game's narrative is included.
        """
        if fast_downward_missing:
            msg = "PddlEnv requires fast_downward to be installed. Try running\npip install textworld[pddl]"
            raise ImportError(msg)

        super().__init__(infos)
        self.downward_lib = fast_downward.load_lib()

    def _get_entity_infos(self):
        entity_infos = {v.name: EntityInfo(v.name, v.type) for fact in self._pddl_state.facts for v in fact.arguments}

        for info in entity_infos.values():
            info.name = info.id
            info.definite = "the"
            info.indefinite = "a"

        return entity_infos

    def load(self, filename_or_data: Union[str, Mapping]) -> None:
        try:
            data = json.load(open(filename_or_data))
            self._game_file = filename_or_data
        except TypeError:
            data = filename_or_data
            self._game_file = None

        self._game_data = data
        self._logic = pddl_logic.GameLogic(domain=self._game_data["pddl_domain"], grammar=self._game_data["grammar"])
        self._pddl_state = pddl_logic.PddlState(self.downward_lib, self._game_data["pddl_problem"], self._logic)
        self._entity_infos = self._get_entity_infos()

    def _get_human_readable_fact(self, fact: Proposition) -> Proposition:
        def _get_name(info):
            return info.name if info.name else info.id

        arguments = [Variable(_get_name(self._entity_infos[var.name]), var.type) for var in fact.arguments]
        return Proposition(fact.name, arguments)

    def _gather_infos(self):
        self.state["command_templates"] = sorted(set(action.template for action in self._logic.actions.values()))

        self.state["won"] = self._pddl_state.check_goal()
        self.state["lost"] = False  # TODO: ask planner if the game is lost.

        self.state["_entity_infos"] = dict(self._entity_infos)
        self.state["_facts"] = list(self._pddl_state.facts)
        if self.request_infos.facts:
            self.state["facts"] = list(map(self._get_human_readable_fact, self.state["_facts"]))

        self.state["last_action"] = None
        self.state["_last_action"] = self._last_action
        self.state["_valid_actions"] = list(self._pddl_state.all_applicable_actions())

        mapping = {k: info.name for k, info in self._entity_infos.items()}
        self.state["_valid_commands"] = []
        for action in self.state["_valid_actions"]:
            context = {
                "state": self._pddl_state,
                "facts": list(self._pddl_state.facts),
                "variables": {ph.name: self._entity_infos[var.name] for ph, var in action.mapping.items()},
                "mapping": action.mapping,
                "entity_infos": self._entity_infos,
            }
            action.command_template = self._logic.grammar.derive(action.command_template, context)
            self.state["_valid_commands"].append(action.format_command(mapping))

        # To guarantee the order from one execution to another, we sort the commands.
        # Remove any potential duplicate commands (they would lead to the same result anyway).
        self.state["admissible_commands"] = sorted(set(self.state["_valid_commands"]))

        if self.request_infos.moves:
            self.state["moves"] = self._moves

        if self.request_infos.policy_commands:
            self.state["policy_commands"] = self._pddl_state.replan(self._entity_infos)

    def reset(self):
        self._pddl_state = pddl_logic.PddlState(self.downward_lib, self._game_data["pddl_problem"], self._logic)

        self.prev_state = None
        self.state = GameState()
        self._last_action = None
        self._moves = 0

        context = {
            "state": self._pddl_state,
            "facts": list(self._pddl_state.facts),
            "variables": {},
            "mapping": {},
            "entity_infos": self._entity_infos,
        }

        self.state.feedback = self._logic.grammar.derive("#intro#", context)
        self.state.raw = self.state.feedback
        self._gather_infos()

        if "walkthrough" in self.request_infos.extras:
            self.state["extra.walkthrough"] = self._pddl_state.replan(self._entity_infos)

        return self.state

    def step(self, command: str):
        command = command.strip()
        self.prev_state = self.state

        self.state = GameState()
        self.state.last_command = command

        self._last_action = None
        try:
            # Find the action corresponding to the command.
            idx = self.prev_state["_valid_commands"].index(command)
            self._last_action = self.prev_state["_valid_actions"][idx]
            # An action that affects the state of the game.
            self._pddl_state.apply(self._last_action)

            context = {
                "state": self._pddl_state,
                "facts": list(self._pddl_state.facts),
                "variables": {ph.name: self._entity_infos[var.name] for ph, var in self._last_action.mapping.items()},
                "mapping": self._last_action.mapping,
                "entity_infos": self._entity_infos,
            }

            self.state.feedback = self._logic.grammar.derive(self._last_action.feedback_rule, context)
            self._moves += 1
        except ValueError:
            # We assume nothing happened in the game.
            self.state.feedback = "Nothing happens."

        self.state.raw = self.state.feedback
        self._gather_infos()
        self.state["score"] = 1 if self.state["won"] else 0
        self.state["done"] = self.state["won"] or self.state["lost"]
        return self.state, self.state["score"], self.state["done"]



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT license.

from typing import Optional

import textworld

import textworld.envs
from textworld.core import EnvInfos, Wrapper


class GenericEnvironment(Wrapper):

    def __init__(self, request_infos: Optional[EnvInfos] = None) -> None:
        super().__init__()
        self.request_infos = request_infos
        self._last_backend = None

    def load(self, gamefile: str) -> None:
        backend = textworld.envs._guess_backend(gamefile)
        if self._last_backend != backend:
            self._wrap(textworld.start(gamefile, self.request_infos))
            self._last_backend = backend
        else:
            self._wrapped_env.load(gamefile)



#!/usr/bin/env python

# CAVEAT UTILITOR
#
# This file was automatically generated by TatSu.
#
#    https://pypi.python.org/pypi/tatsu/
#
# Any changes you make to it will be overwritten the next time
# the file is generated.

from __future__ import annotations

import sys

from tatsu.buffering import Buffer
from tatsu.parsing import Parser
from tatsu.parsing import tatsumasu
from tatsu.parsing import leftrec, nomemo, isname # noqa
from tatsu.infos import ParserConfig
from tatsu.util import re, generic_main  # noqa


KEYWORDS = {}  # type: ignore


class TextGrammarBuffer(Buffer):
    def __init__(self, text, /, config: ParserConfig = None, **settings):
        config = ParserConfig.new(
            config,
            owner=self,
            whitespace=re.compile('[\\t ]+'),
            nameguard=None,
            comments_re=None,
            eol_comments_re='^(#.*|\\s*)\\n',
            ignorecase=False,
            namechars='',
            parseinfo=False,
        )
        config = config.replace(**settings)
        super().__init__(text, config=config)


class TextGrammarParser(Parser):
    def __init__(self, /, config: ParserConfig = None, **settings):
        config = ParserConfig.new(
            config,
            owner=self,
            whitespace=re.compile('[\\t ]+'),
            nameguard=None,
            comments_re=None,
            eol_comments_re='^(#.*|\\s*)\\n',
            ignorecase=False,
            namechars='',
            parseinfo=False,
            keywords=KEYWORDS,
            start='start',
        )
        config = config.replace(**settings)
        super().__init__(config=config)

    @tatsumasu()
    def _start_(self):  # noqa
        self._grammar_()

    @tatsumasu()
    def _symbol_(self):  # noqa
        self._pattern('[\\w()/!<>-]+')

    @tatsumasu()
    def _literal_(self):  # noqa
        self._pattern('((?:[^;|<>\\n\\[\\]()]|\\[[^\\[\\]]*\\]|\\([^()]*\\))+(?<!\\s))')

    @tatsumasu('Literal')
    def _literalAlternative_(self):  # noqa
        self._literal_()
        self.name_last_node('value')

    @tatsumasu('AdjectiveNoun')
    def _adjectiveNoun_(self):  # noqa
        self._literal_()
        self.name_last_node('adjective')
        self._token('|')
        self._literal_()
        self.name_last_node('noun')

        self._define(
            ['adjective', 'noun'],
            []
        )

    @tatsumasu()
    def _entity_(self):  # noqa
        with self._choice():
            with self._option():
                self._adjectiveNoun_()
            with self._option():
                self._literalAlternative_()
            self._error(
                'expecting one of: '
                '((?:[^;|<>\\n\\[\\]()]|\\[[^\\[\\]]*\\]|\\([^()]'
                '*\\))+(?<!\\s)) <adjectiveNoun> <literal>'
                '<literalAlternative>'
            )

    @tatsumasu('Match')
    def _match_(self):  # noqa
        self._entity_()
        self.name_last_node('lhs')
        self._token('<->')
        self._entity_()
        self.name_last_node('rhs')

        self._define(
            ['lhs', 'rhs'],
            []
        )

    @tatsumasu()
    def _alternative_(self):  # noqa
        with self._choice():
            with self._option():
                self._match_()
            with self._option():
                self._entity_()
            with self._option():
                self._void()
            self._error(
                'expecting one of: '
                '((?:[^;|<>\\n\\[\\]()]|\\[[^\\[\\]]*\\]|\\([^()]'
                '*\\))+(?<!\\s)) <adjectiveNoun> <entity>'
                '<literal> <literalAlternative> <match>'
            )

    @tatsumasu()
    def _alternatives_(self):  # noqa

        def sep0():
            self._token(';')

        def block0():
            self._alternative_()
        self._positive_gather(block0, sep0)

    @tatsumasu('ProductionRule')
    def _productionRule_(self):  # noqa
        self._symbol_()
        self.name_last_node('symbol')
        self._token(':')
        self._alternatives_()
        self.name_last_node('alternatives')
        with self._group():
            with self._choice():
                with self._option():
                    self._token('\n')
                with self._option():
                    self._check_eof()
                self._error(
                    'expecting one of: '
                    "'\\n'"
                )

        self._define(
            ['alternatives', 'symbol'],
            []
        )

    @tatsumasu('TextGrammar')
    def _grammar_(self):  # noqa

        def block1():
            self._productionRule_()
        self._closure(block1)
        self.name_last_node('rules')
        self._check_eof()

        self._define(
            ['rules'],
            []
        )


class TextGrammarSemantics:
    def start(self, ast):  # noqa
        return ast

    def symbol(self, ast):  # noqa
        return ast

    def literal(self, ast):  # noqa
        return ast

    def literalAlternative(self, ast):  # noqa
        return ast

    def adjectiveNoun(self, ast):  # noqa
        return ast

    def entity(self, ast):  # noqa
        return ast

    def match(self, ast):  # noqa
        return ast

    def alternative(self, ast):  # noqa
        return ast

    def alternatives(self, ast):  # noqa
        return ast

    def productionRule(self, ast):  # noqa
        return ast

    def grammar(self, ast):  # noqa
        return ast


def main(filename, **kwargs):
    if not filename or filename == '-':
        text = sys.stdin.read()
    else:
        with open(filename) as f:
            text = f.read()
    parser = TextGrammarParser()
    return parser.parse(
        text,
        filename=filename,
        **kwargs
    )


if __name__ == '__main__':
    import json
    from tatsu.util import asjson

    ast = generic_main(main, TextGrammarParser, name='TextGrammar')
    data = asjson(ast)
    print(json.dumps(data, indent=2))



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT license.


import textworld
from textworld import g_rng
from textworld.utils import make_temp_directory


def test_making_game_with_names_to_exclude():
    g_rng.set_seed(42)

    with make_temp_directory(prefix="test_render_wrapper") as tmpdir:
        options = textworld.GameOptions()
        options.path = tmpdir
        options.nb_rooms = 2
        options.nb_objects = 20
        options.chaining.max_depth = 3
        options.chaining.max_breadth = 2
        options.seeds = 123
        game_file1, game1 = textworld.make(options)

        options2 = options.copy()
        game1_objects_names = [info.name for info in game1.infos.values() if info.name is not None]
        options2.grammar.names_to_exclude = game1_objects_names
        game_file2, game2 = textworld.make(options2)
        game2_objects_names = [info.name for info in game2.infos.values() if info.name is not None]
        assert len(set(game1_objects_names) & set(game2_objects_names)) == 0


def test_making_game_is_reproducible_with_seed():
    with make_temp_directory(prefix="test_render_wrapper") as tmpdir:
        options = textworld.GameOptions()
        options.path = tmpdir
        options.nb_rooms = 2
        options.nb_objects = 20
        options.chaining.max_depth = 3
        options.chaining.max_breadth = 2
        options.seeds = 123

        game_file1, game1 = textworld.make(options)
        options2 = options.copy()
        game_file2, game2 = textworld.make(options2)
        assert game_file1 == game_file2
        assert game1 == game2
        # Make sure they are not the same Python objects.
        assert id(game1) != id(game2)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT license.

import os
import glob
from subprocess import check_call, CalledProcessError
from os.path import join as pjoin
import textwrap

import textworld
import textworld.agents
import textworld.challenges
from textworld.utils import make_temp_directory


def test_making_a_custom_game():
    with make_temp_directory(prefix="test_tw-make") as tmpdir:
        output_folder = pjoin(tmpdir, "gen_games")
        game_file = pjoin(output_folder, "game_1234.ulx")
        command = ["tw-make", "custom", "--seed", "1234", "--output", game_file, "--silent"]
        assert check_call(command) == 0

        assert os.path.isdir(output_folder)
        assert os.path.isfile(game_file)

        # Solve the game using WalkthroughAgent.
        agent = textworld.agents.WalkthroughAgent()
        textworld.play(game_file, agent=agent, silent=True)

    with make_temp_directory(prefix="test_tw-make") as tmpdir:
        output_folder = pjoin(tmpdir, "gen_games")
        game_file = pjoin(output_folder, "game_1234")  # Default extension is .z8
        command = ["tw-make", "custom", "--seed", "1234", "--output", game_file, "--silent"]
        assert check_call(command) == 0

        assert os.path.isdir(output_folder)
        assert os.path.isfile(game_file + ".z8")

        # Solve the game using WalkthroughAgent.
        agent = textworld.agents.WalkthroughAgent()
        textworld.play(game_file + ".z8", agent=agent, silent=True)

    with make_temp_directory(prefix="test_tw-make") as tmpdir:
        output_folder = pjoin(tmpdir, "gen_games", "")
        command = ["tw-make", "custom", "--seed", "1234", "--output", output_folder, "--silent"]
        assert check_call(command) == 0

        assert os.path.isdir(output_folder)
        game_file = glob.glob(pjoin(output_folder, "*.z8"))[0]

        # Solve the game using WalkthroughAgent.
        agent = textworld.agents.WalkthroughAgent()
        textworld.play(game_file, agent=agent, silent=True)

    with make_temp_directory(prefix="test_tw-make") as tmpdir:
        output_folder = pjoin(tmpdir, "gen_games")
        command = ["tw-make", "custom", "--seed", "1234", "--output", output_folder, "--silent"]
        assert check_call(command) == 0

        assert os.path.isfile(output_folder + ".z8")

        # Solve the game using WalkthroughAgent.
        agent = textworld.agents.WalkthroughAgent()
        textworld.play(output_folder + ".z8", agent=agent, silent=True)


def test_making_challenge_game():
    settings = {
        "tw-treasure_hunter": [["--level", "5"]],
        "tw-coin_collector": [["--level", "5"]],
        "tw-simple": [["--rewards", "dense", "--goal", "brief"]],
        "tw-cooking": [["--recipe", "2", "--take", "1", "--cook", "--split", "valid"],
                       ["--recipe", "2", "--take", "1", "--cook", "--drop", "--split", "valid"]],
    }
    with make_temp_directory(prefix="test_tw-challenge") as tmpdir:
        for challenge in textworld.challenges.CHALLENGES:
            for i, params in enumerate(settings[challenge]):
                output_folder = pjoin(tmpdir, "gen_games")
                game_file = pjoin(output_folder, challenge + "_{}".format(i) + ".ulx")
                command = ["tw-make", challenge, "--seed", "1234", "--output", game_file, "--silent"] + params
                assert check_call(command) == 0

                assert os.path.isdir(output_folder)
                assert os.path.isfile(game_file)

                # Solve the game using WalkthroughAgent.
                agent = textworld.agents.WalkthroughAgent()
                textworld.play(game_file, agent=agent, silent=True)


def test_making_a_game_using_basic_theme():
    for i in range(10):  # Try a few different games.
        with make_temp_directory(prefix="test_tw-make") as tmpdir:
            output_folder = pjoin(tmpdir, "gen_games")
            game_file = pjoin(output_folder, "game_1234.ulx")
            command = ["tw-make", "custom", "--theme", "basic", "--seed", str(i), "--output", game_file, "--silent"]
            assert check_call(command) == 0

            # Solve the game using WalkthroughAgent.
            agent = textworld.agents.WalkthroughAgent()
            textworld.play(game_file, agent=agent, silent=True)


def test_third_party():
    with make_temp_directory(prefix="test_tw-make_third_party") as tmpdir:
        challenge_py = pjoin(tmpdir, "my_challenge.py")
        with open(challenge_py, "w") as f:
            f.write(textwrap.dedent("""\
            import argparse
            from typing import Mapping, Optional

            import textworld
            from textworld.challenges import register

            from textworld import Game, GameOptions


            def build_argparser(parser=None):
                parser = parser or argparse.ArgumentParser()

                group = parser.add_argument_group('This challenge settings')
                group.add_argument("--nb-locations", required=True, type=int,
                                   help="Number of locations in the game.")

                return parser


            def make_game(settings: Mapping[str, str], options: Optional[GameOptions] = None) -> Game:
                options = options or GameOptions
                options.nb_rooms = settings["nb_locations"]
                game = textworld.generator.make_game(options)
                return game


            # Register this new challenge.
            register(name="my-challenge",
                    desc="Generate new challenge game",
                    make=make_game,
                    add_arguments=build_argparser)
            """))

        NB_LOCATIONS = 3

        output_folder = pjoin(tmpdir, "gen_games")
        game_file = pjoin(output_folder, "game_1234.ulx")
        command = ["tw-make", "--third-party", challenge_py, "my-challenge", "--seed", "1234",
                   "--output", game_file, "--silent"]
        try:
            check_call(command)
        except CalledProcessError as e:
            assert e.returncode == 2  # Missing the --nb-locations argument.
        else:
            assert False, "tw-make should have failed."

        command += ["--nb-locations", str(NB_LOCATIONS)]
        assert check_call(command) == 0

        game = textworld.Game.load(game_file.replace(".ulx", ".json"))
        assert len(game.world.rooms) == NB_LOCATIONS



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT license.

import os
from os.path import join as pjoin
from subprocess import check_output

from textworld.utils import make_temp_directory

SCRIPTS_PATH = os.path.abspath(pjoin(__file__, "..", "..", "scripts"))


def test_sample_quests():
    with make_temp_directory(prefix="test_sample_quests") as tmpdir:
        game_file = pjoin(tmpdir, "game.ulx")
        command = ["tw-make", "custom", "--seed", "20181004", "--output", game_file]
        check_output(command).decode()

        script = pjoin(SCRIPTS_PATH, "sample_quests.py")
        command = ["python", script, "--nb-quests", "10", "--quest-length", "10",
                   "--quest-breadth", "5", "--output", tmpdir, game_file]
        stdout = check_output(command).decode()
        assert len(stdout) > 0
        assert os.path.isfile(pjoin(tmpdir, "sample_world.png"))
        assert os.path.isfile(pjoin(tmpdir, "sample_tree.svg"))
        assert os.path.isfile(pjoin(tmpdir, "sample_graph.svg"))



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT license.


import glob
import shutil
import tempfile
import unittest
from os.path import join as pjoin

import numpy as np

import textworld
import textworld.agents
from textworld.utils import make_temp_directory


class TestIntegration(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.tmpdir = pjoin(tempfile.mkdtemp(prefix="test_textworld"), "")
        options = textworld.GameOptions()
        options.path = cls.tmpdir
        options.nb_rooms = 5
        options.nb_objects = 10
        options.quest_length = 10
        options.seeds = 1234
        cls.game_file, cls.game = textworld.make(options)
        options.file_ext = ".z8"
        cls.game_file_z8, _ = textworld.make(options)

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(cls.tmpdir)

    def setUp(self):
        self.before_tw = glob.glob(pjoin(tempfile.gettempdir(), "tw_*"))
        self.before_mlglk = glob.glob(pjoin(tempfile.gettempdir(), "mlglk_*"))

    def tearDown(self):
        # Check for file leaks.
        after_tw = glob.glob(pjoin(tempfile.gettempdir(), "tw_*"))
        after_mlglk = glob.glob(pjoin(tempfile.gettempdir(), "mlglk_*"))
        assert set(after_tw) == set(self.before_tw)
        assert set(after_mlglk) == set(self.before_mlglk)

    def test_100_sequential_runs(self):
        for gamefile in [self.game_file, self.game_file_z8]:
            for _ in range(100):
                env = textworld.start(gamefile)
                env.reset()
                game_state, reward, done = env.step('take inventory')
                self.assertIsNotNone(game_state, "Checking gamestate is not None")
                self.assertIsNotNone(reward, "Checking reward is not None")
                self.assertFalse(done, "Checking we don't finish the game by looking at our stuff")
                env.close()

    def test_simultaneous_runs(self):
        envs = []
        for gamefile in [self.game_file, self.game_file_z8]:
            for _ in range(100):
                env = textworld.start(gamefile)
                env.reset()
                envs.append(env)

            game_state, reward, done = envs[-1].step('take inventory')
            self.assertIsNotNone(game_state, "Checking gamestate is not None")
            self.assertIsNotNone(reward, "Checking reward is not None")
            self.assertFalse(done, "Checking we don't finish the game by looking at our stuff")

        for env in envs:
            env.close()

    def test_game_random_agent(self):
        for gamefile in [self.game_file, self.game_file_z8]:
            env = textworld.start(gamefile)
            agent = textworld.agents.RandomCommandAgent()
            agent.reset(env)
            game_state = env.reset()

            reward = 0
            done = False
            for _ in range(5):
                command = agent.act(game_state, reward, done)
                game_state, reward, done = env.step(command)

    def test_game_walkthrough_agent(self):
        agent = textworld.agents.WalkthroughAgent()

        for gamefile in [self.game_file, self.game_file_z8]:

            env = textworld.start(gamefile)

            agent.reset(env)
            game_state = env.reset()
            commands = game_state.get("extra.walkthrough")

            reward = 0
            done = False
            for walkthrough_command in commands:
                self.assertFalse(done, 'walkthrough finished game too early')
                command = agent.act(game_state, reward, done)
                self.assertEqual(walkthrough_command, command, "Walkthrough agent issued unexpected command")
                game_state, reward, done = env.step(command)

            self.assertTrue(done, 'Walkthrough did not finish the game')


def test_playing_generated_games():
    NB_GAMES = 10
    rng = np.random.RandomState(1234)
    for _ in range(NB_GAMES):

        # Sample game specs.
        world_size = rng.randint(1, 10)
        nb_objects = rng.randint(0, 20)
        quest_depth = rng.randint(2, 5)
        quest_breadth = rng.randint(3, 7)
        game_seed = rng.randint(0, 65365)

        with make_temp_directory(prefix="test_play_generated_games") as tmpdir:
            options = textworld.GameOptions()
            options.path = tmpdir
            options.nb_rooms = world_size
            options.nb_objects = nb_objects
            options.chaining.max_depth = quest_depth
            options.chaining.max_breadth = quest_breadth
            options.seeds = game_seed
            game_file, _ = textworld.make(options)

            # Solve the game using WalkthroughAgent.
            agent = textworld.agents.WalkthroughAgent()
            textworld.play(game_file, agent=agent, silent=True)

            # Play the game using RandomAgent and make sure we can always finish the
            # game by following the winning policy.
            env = textworld.start(game_file)
            env.request_infos.policy_commands = True
            env.request_infos.game = True

            agent = textworld.agents.RandomCommandAgent()
            agent.reset(env)

            env.seed(4321)
            game_state = env.reset()

            max_steps = 100
            reward = 0
            done = False
            for step in range(max_steps):
                command = agent.act(game_state, reward, done)
                game_state, reward, done = env.step(command)

                if done:
                    assert game_state._winning_policy is None
                    game_state, reward, done = env.reset(), 0, False

                # Make sure the game can still be solved.
                winning_policy = game_state._winning_policy
                assert len(winning_policy) > 0
                assert game_state._game_progression.state.is_sequence_applicable(winning_policy)




import json
import textwrap

from collections import Counter, defaultdict

from tatsu.model import NodeWalker

import fast_downward

import textworld.logic.model
from textworld.utils import check_flag
from textworld.logic import Proposition, Variable, Placeholder

from textworld.envs.pddl.textgen import ContextSensitiveGrammar
from textworld.envs.pddl.logic.model import PddlLogicModelBuilderSemantics
from textworld.envs.pddl.logic.parser import PddlLogicParser


class _ModelConverter(NodeWalker):
    """
    Converts TatSu model objects to our types.
    """

    def __init__(self, logic=None):
        super().__init__()
        self._cache = {}
        self._logic = logic

    def _unescape(self, string):
        # Strip quotation marks
        return string[1:-1]

    def _unescape_block(self, string):
        # Strip triple quotation marks and dedent
        string = string[3:-3]
        return textwrap.dedent(string)

    def walk_list(self, l):
        return [self.walk(node) for node in l]

    def walk_ActionTypeNode(self, node):
        action = Action(
            name=node.name,
            template=self._unescape(node.template.template),
            feedback_rule=self._unescape(node.feedback.name),
            pddl=self._unescape_block(node.pddl.code) if node.pddl else "",
            grammar=self._unescape_block(node.grammar.code) if node.grammar else "{}"
        )

        return action

    def walk_PddlDocumentNode(self, node):
        actions = {}
        grammar = {}
        for part in node.parts:
            if isinstance(part, textworld.envs.pddl.logic.model.ActionTypeNode):
                action = self.walk(part)
                actions[action.name.lower()] = action
                grammar_data = "\n".join(line for line in action.grammar.split("\n") if not line.lstrip().startswith("#"))
                grammar.update(json.loads(grammar_data))
            elif isinstance(part, textworld.envs.pddl.logic.model.ActionGrammarNode):
                grammar.update(json.loads(self._unescape_block(part.code)))

        grammar = ContextSensitiveGrammar.parse(json.dumps(grammar))
        return actions, grammar


_PARSER = PddlLogicParser(semantics=PddlLogicModelBuilderSemantics(), parseinfo=True)


def _parse_and_convert(*args, **kwargs):
    model = _PARSER.parse(*args, **kwargs)
    return _ModelConverter().walk(model)


class Action:
    def __init__(self, name, template, pddl, grammar, feedback_rule):
        self.name = name
        self.feedback_rule = feedback_rule
        self.template = template
        self.pddl = pddl
        self.grammar = grammar


class GameLogic:
    """
    The logic for a game (types, rules, grammar, etc.).
    """

    def __init__(self, domain, grammar):
        self.domain = domain
        self.grammar = ContextSensitiveGrammar()
        self.actions = {}
        self.types = textworld.logic.TypeHierarchy()

        # Load grammar's content
        actions, grammar = _parse_and_convert(grammar, rule_name="pddlStart")
        self.actions.update(actions)
        self.grammar.update(grammar)


class Atom(fast_downward.Atom):
    """
    Customized fast_downward.Atom that can easily export Atom objects as Propositions.
    """

    def _get_var_name(self, value):
        if value.lower() in ("i", "p"):
            return value.upper()

        return value

    def _get_var_type(self, value):
        if value.lower() in ("i", "p"):
            return value.upper()

        return value[0]

    @property
    def as_fact(self) -> Proposition:
        atom_type, rest = self.name.split(" ", 1)
        name, args = rest.split("(", 1)
        args = args[:-1].split(", ")
        arguments = [Variable(self._get_var_name(arg), self._get_var_type(arg)) for arg in args if arg]
        if atom_type == "NegatedAtom":
            name = "not_" + name

        return Proposition(name, arguments)

    def get_fact(self, name2type={}) -> Proposition:
        atom_type, rest = self.name.split(" ", 1)
        name, args = rest.split("(", 1)
        args = args[:-1].split(", ")
        arguments = [Variable(self._get_var_name(arg), name2type[arg]) for arg in args if arg]
        if atom_type == "NegatedAtom":
            name = "not_" + name

        return Proposition(name, arguments)


class PddlState(textworld.logic.State):
    """
    The current state of a game.
    """

    def __init__(self, downward_lib, pddl_problem: str, logic: GameLogic):
        """
        Create a State from PDDL problem.

        Parameters
        ----------
        downward_lib:
        pddl_problem:
        logic :
            The logic for this state's game.
        """
        self._facts = defaultdict(set)
        self._vars_by_name = {}
        self._vars_by_type = defaultdict(set)
        self._var_counts = Counter()

        self._logic = logic
        self.downward_lib = downward_lib

        # problem
        self.pddl_problem = pddl_problem

        # Load domain + problem.
        verbose = check_flag("TW_PDDL_DEBUG")
        self.task, self.sas = fast_downward.pddl2sas(logic.domain, pddl_problem, verbose=verbose)
        _, self.sas_replan = fast_downward.pddl2sas(logic.domain, pddl_problem, verbose=verbose, optimize=True)

        self._actions = {a.name: a for a in self.task.actions}

        # Import types from fastdownward
        for type_ in self.task.types:
            try:
                type_ = textworld.logic.Type(type_.name, type_.supertype_names)
                self._logic.types.add(type_)
            except ValueError:
                pass

        self.downward_lib.load_sas(self.sas.encode('utf-8'))
        self.downward_lib.load_sas_replan(self.sas_replan.encode('utf-8'))

        self.name2type = {o.name: o.type_name for o in self.task.objects}

        def _atom2proposition(atom):
            if isinstance(atom, fast_downward.translate.pddl.conditions.Atom):
                if atom.predicate == "=":
                    return None

                return Proposition(atom.predicate, [Variable(arg, self.name2type[arg]) for arg in atom.args])

            elif isinstance(atom, fast_downward.translate.pddl.f_expression.Assign):
                if atom.fluent.symbol == "total-cost":
                    return None

                name = "{}".format(atom.expression.value)
                if str.isdigit(name):  # Discard distance relations.
                    return None

                return Proposition(name, [Variable(arg, self.name2type[arg]) for arg in atom.fluent.args])

        facts = [_atom2proposition(atom) for atom in self.task.init]
        facts = list(filter(None, facts))
        self.add_facts(facts)

        state_size = self.downward_lib.get_state_size()
        atoms = (Atom * state_size)()
        self.downward_lib.get_state(atoms)
        facts = [atom.get_fact(self.name2type) for atom in atoms]
        facts = [fact for fact in facts if not fact.is_negation]
        self.add_facts(facts)

    def all_applicable_actions(self):
        operator_count = self.downward_lib.get_applicable_operators_count()

        operators = (fast_downward.Operator * operator_count)()
        self.downward_lib.get_applicable_operators(operators)
        self._operators = {op.id: op for op in operators}

        actions = []
        seen_operators = set()
        for operator in operators:
            if operator.name in seen_operators:
                continue

            seen_operators.add(operator.name)

            splits = operator.name.split()
            name, arguments = splits[0], splits[1:]

            action = textworld.logic.Action(name=name, preconditions=[], postconditions=[])
            action.id = operator.id
            action.mapping = {Placeholder(p.name.strip("?"), p.type_name): Variable(arg, p.type_name)
                              for p, arg in zip(self._actions[name].parameters, arguments)}
            action.command_template = self._logic.actions[name].template
            action.feedback_rule = self._logic.actions[name].feedback_rule
            actions.append(action)

        return actions

    def apply(self, action):
        # TODO: convert textworld.logic.Action into operator id.
        op = self._operators[action.id]  # HACK: assume action is operator id for now.

        effects = (Atom * op.nb_effect_atoms)()
        self.downward_lib.apply_operator(op.id, effects)

        # Update facts
        changes = []
        for effect in effects:
            prop = effect.get_fact(self.name2type)
            changes.append(prop)
            self.remove_fact(prop.negate())
            self.add_fact(prop)

        return changes

    def check_goal(self):
        return self.downward_lib.check_goal()

    def replan(self, infos):
        if not self.downward_lib.replan(check_flag("TW_PDDL_DEBUG")):
            return []

        operators = (fast_downward.Operator * self.downward_lib.get_last_plan_length())()
        self.downward_lib.get_last_plan(operators)
        plan = [op.name for op in operators]
        templated_actions = self.plan_to_templated_actions(plan, infos)
        return templated_actions

    def plan_to_templated_actions(self, plan, infos):
        templated_actions = []
        for step in plan:
            components = step.split()
            action = components[0]
            action_params = [a.name.replace('?', '') for a in self._actions[action].parameters]
            param_values = {str(param): infos[step.split()[i + 1]].name
                            for i, param in enumerate(action_params)}
            template = self._logic.actions[action].template if action != 'gotolocation' else "go to {r}"
            template = template.format(**param_values)
            templated_actions.append(template)

        return templated_actions



#!/usr/bin/env python

# CAVEAT UTILITOR
#
# This file was automatically generated by TatSu.
#
#    https://pypi.python.org/pypi/tatsu/
#
# Any changes you make to it will be overwritten the next time
# the file is generated.

from __future__ import annotations

import sys

from tatsu.buffering import Buffer
from tatsu.parsing import Parser
from tatsu.parsing import tatsumasu
from tatsu.parsing import leftrec, nomemo, isname # noqa
from tatsu.infos import ParserConfig
from tatsu.util import re, generic_main  # noqa


KEYWORDS = {}  # type: ignore


class PddlLogicBuffer(Buffer):
    def __init__(self, text, /, config: ParserConfig = None, **settings):
        config = ParserConfig.new(
            config,
            owner=self,
            whitespace=None,
            nameguard=None,
            comments_re=None,
            eol_comments_re='#.*?$',
            ignorecase=False,
            namechars='',
            parseinfo=False,
        )
        config = config.replace(**settings)
        super().__init__(text, config=config)


class PddlLogicParser(Parser):
    def __init__(self, /, config: ParserConfig = None, **settings):
        config = ParserConfig.new(
            config,
            owner=self,
            whitespace=None,
            nameguard=None,
            comments_re=None,
            eol_comments_re='#.*?$',
            ignorecase=False,
            namechars='',
            parseinfo=False,
            keywords=KEYWORDS,
            start='start',
        )
        config = config.replace(**settings)
        super().__init__(config=config)

    @tatsumasu()
    def _start_(self):  # noqa
        self._pddlDocument_()

    @tatsumasu()
    def _str_(self):  # noqa
        self._pattern('"[^"]*"')

    @tatsumasu()
    def _strBlock_(self):  # noqa
        self._pattern('"""(?:.|\\n)*?"""')

    @tatsumasu()
    def _name_(self):  # noqa
        self._pattern('[\\w/-]+')

    @tatsumasu('ActionTemplateNode')
    def _template_(self):  # noqa
        self._token('template')
        self._token('::')
        self._str_()
        self.name_last_node('template')
        self._token(';')

        self._define(
            ['template'],
            []
        )

    @tatsumasu('ActionFeedbackNode')
    def _feedback_(self):  # noqa
        self._token('feedback')
        self._token('::')
        self._str_()
        self.name_last_node('name')
        self._token(';')

        self._define(
            ['name'],
            []
        )

    @tatsumasu('ActionPddlNode')
    def _pddl_(self):  # noqa
        self._token('pddl')
        self._token('::')
        self._strBlock_()
        self.name_last_node('code')
        self._token(';')

        self._define(
            ['code'],
            []
        )

    @tatsumasu('ActionGrammarNode')
    def _grammar_(self):  # noqa
        self._token('grammar')
        self._token('::')
        self._strBlock_()
        self.name_last_node('code')
        self._token(';')

        self._define(
            ['code'],
            []
        )

    @tatsumasu('ActionTypeNode')
    def _actionType_(self):  # noqa
        self._token('action')
        self._name_()
        self.name_last_node('name')
        self._token('{')

        def block1():
            with self._group():
                with self._choice():
                    with self._option():
                        self._template_()
                        self.name_last_node('template')
                    with self._option():
                        self._feedback_()
                        self.name_last_node('feedback')
                    with self._option():
                        self._pddl_()
                        self.name_last_node('pddl')
                    with self._option():
                        self._grammar_()
                        self.name_last_node('grammar')
                    self._error(
                        'expecting one of: '
                        '<feedback> <grammar> <pddl> <template>'
                    )
        self._closure(block1)
        self._token('}')

        self._define(
            ['feedback', 'grammar', 'name', 'pddl', 'template'],
            []
        )

    @tatsumasu('PddlDocumentNode')
    def _pddlDocument_(self):  # noqa

        def block1():
            with self._group():
                with self._choice():
                    with self._option():
                        self._actionType_()
                    with self._option():
                        self._grammar_()
                    self._error(
                        'expecting one of: '
                        '<actionType> <grammar>'
                    )
        self._closure(block1)
        self.name_last_node('parts')
        self._check_eof()

        self._define(
            ['parts'],
            []
        )

    @tatsumasu()
    def _pddlStart_(self):  # noqa
        self._pddlDocument_()


class PddlLogicSemantics:
    def start(self, ast):  # noqa
        return ast

    def str(self, ast):  # noqa
        return ast

    def strBlock(self, ast):  # noqa
        return ast

    def name(self, ast):  # noqa
        return ast

    def template(self, ast):  # noqa
        return ast

    def feedback(self, ast):  # noqa
        return ast

    def pddl(self, ast):  # noqa
        return ast

    def grammar(self, ast):  # noqa
        return ast

    def actionType(self, ast):  # noqa
        return ast

    def pddlDocument(self, ast):  # noqa
        return ast

    def pddlStart(self, ast):  # noqa
        return ast


def main(filename, **kwargs):
    if not filename or filename == '-':
        text = sys.stdin.read()
    else:
        with open(filename) as f:
            text = f.read()
    parser = PddlLogicParser()
    return parser.parse(
        text,
        filename=filename,
        **kwargs
    )


if __name__ == '__main__':
    import json
    from tatsu.util import asjson

    ast = generic_main(main, PddlLogicParser, name='PddlLogic')
    data = asjson(ast)
    print(json.dumps(data, indent=2))



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT license.

import re
import os
import glob
import shutil
import unittest
import tempfile
from os.path import join as pjoin
from subprocess import check_output

import textworld


class TestTwExtract(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.tmpdir = pjoin(tempfile.mkdtemp(prefix="test_tw_extract"), "")
        options = textworld.GameOptions()
        options.path = cls.tmpdir
        options.nb_rooms = 5
        options.nb_objects = 10
        options.quest_length = 5
        options.quest_breadth = 2
        options.seeds = 1234
        cls.game_file1, cls.game1 = textworld.make(options)
        options.seeds = 12345
        options.file_ext = ".z8"
        cls.game_file2, cls.game2 = textworld.make(options)

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(cls.tmpdir)

    def tearDown(self):
        for f in glob.glob(pjoin(self.tmpdir, "*.txt")):
            os.remove(f)

    def test_extract_vocab(self):
        outfile = pjoin(self.tmpdir, "vocab.txt")
        command = ["tw-extract", "vocab", self.game_file1, self.game_file2, "--output", outfile]
        stdout = check_output(command).decode()
        assert os.path.isfile(outfile)
        nb_words = len(open(outfile).readlines())
        assert "Found {}".format(nb_words) in stdout

    def test_extract_vocab_theme(self):
        outfile = pjoin(self.tmpdir, "vocab.txt")
        command = ["tw-extract", "vocab", "--theme", "house", "--output", outfile]
        stdout = check_output(command).decode()
        assert os.path.isfile(outfile)
        assert int(re.findall(r"Found (\d+)", stdout)[0]) > 0

    def test_extract_entities(self):
        outfile = pjoin(self.tmpdir, "entities.txt")
        command = ["tw-extract", "entities", self.game_file1, "--output", outfile]
        stdout = check_output(command).decode()
        assert os.path.isfile(outfile)
        nb_entities = len(open(outfile).readlines())
        assert "Found {}".format(nb_entities) in stdout

    def test_extract_walkthroughs(self):
        outfile = pjoin(self.tmpdir, "walkthroughs.txt")
        command = ["tw-extract", "walkthroughs", self.game_file1, self.game_file2, "--output", outfile]
        stdout = check_output(command).decode()
        assert os.path.isfile(outfile)
        walkthrough1 = " > ".join(self.game1.metadata["walkthrough"])
        walkthrough2 = " > ".join(self.game2.metadata["walkthrough"])
        walkthroughs = open(outfile).readlines()
        assert len(walkthroughs) == 2
        assert walkthrough1 == walkthroughs[0].strip()
        assert walkthrough2 == walkthroughs[1].strip()
        assert "Found {}".format(2) in stdout

        # Simulate game without a walkthrough and the --force argument.
        gamefile = pjoin(self.tmpdir, "game_walkthrough_missing.json")
        game = self.game2.copy()
        del game.metadata["walkthrough"]
        game.save(gamefile)
        command = ["tw-extract", "walkthroughs", gamefile, "--output", outfile, "--force"]
        stdout = check_output(command).decode()
        assert os.path.isfile(outfile)
        walkthroughs = open(outfile).readlines()
        assert len(walkthroughs) == 1
        assert walkthrough2 == walkthroughs[0].strip()

    def test_extract_commands(self):
        outfile = pjoin(self.tmpdir, "commands.txt")
        command = ["tw-extract", "commands", self.game_file1, self.game_file2, "--output", outfile]
        stdout = check_output(command).decode()
        assert os.path.isfile(outfile)
        assert "Found" in stdout

        content = open(outfile).read()
        for name in self.game1.entity_names + self.game2.entity_names:
            assert name in content

        for verb in self.game1.verbs + self.game2.verbs:
            assert verb in content

        # Check inheritance of command templates, e.g., 'examine {t}' -> ['examine {o}', 'examine {k}', 'examine {f}']
        for obj_name in self.game1.objects_names:
            assert "examine {}".format(obj_name) in content



#!/usr/bin/env python

# CAVEAT UTILITOR
#
# This file was automatically generated by TatSu.
#
#    https://pypi.python.org/pypi/tatsu/
#
# Any changes you make to it will be overwritten the next time
# the file is generated.

from __future__ import annotations

from typing import Any
from dataclasses import dataclass

from tatsu.objectmodel import Node
from tatsu.semantics import ModelBuilderSemantics


@dataclass(eq=False)
class ModelBase(Node):
    pass


class GameLogicModelBuilderSemantics(ModelBuilderSemantics):
    def __init__(self, context=None, types=None):
        types = [
            t for t in globals().values()
            if type(t) is type and issubclass(t, ModelBase)
        ] + (types or [])
        super().__init__(context=context, types=types)


@dataclass(eq=False)
class VariableNode(ModelBase):
    name: Any = None
    type: Any = None


@dataclass(eq=False)
class SignatureNode(ModelBase):
    name: Any = None
    types: Any = None


@dataclass(eq=False)
class PropositionNode(ModelBase):
    arguments: Any = None
    name: Any = None


@dataclass(eq=False)
class ActionPreconditionNode(ModelBase):
    condition: Any = None
    preserve: Any = None


@dataclass(eq=False)
class ActionNode(ModelBase):
    name: Any = None
    postconditions: Any = None
    preconditions: Any = None


@dataclass(eq=False)
class PlaceholderNode(ModelBase):
    name: Any = None
    type: Any = None


@dataclass(eq=False)
class PredicateNode(ModelBase):
    name: Any = None
    parameters: Any = None


@dataclass(eq=False)
class RulePreconditionNode(ModelBase):
    condition: Any = None
    preserve: Any = None


@dataclass(eq=False)
class RuleNode(ModelBase):
    name: Any = None
    postconditions: Any = None
    preconditions: Any = None


@dataclass(eq=False)
class AliasNode(ModelBase):
    lhs: Any = None
    rhs: Any = None


@dataclass(eq=False)
class ReverseRuleNode(ModelBase):
    lhs: Any = None
    rhs: Any = None


@dataclass(eq=False)
class PredicatesNode(ModelBase):
    predicates: Any = None


@dataclass(eq=False)
class RulesNode(ModelBase):
    rules: Any = None


@dataclass(eq=False)
class ReverseRulesNode(ModelBase):
    reverse_rules: Any = None


@dataclass(eq=False)
class ConstraintsNode(ModelBase):
    constraints: Any = None


@dataclass(eq=False)
class Inform7TypeNode(ModelBase):
    definition: Any = None
    kind: Any = None


@dataclass(eq=False)
class Inform7PredicateNode(ModelBase):
    predicate: Any = None
    source: Any = None


@dataclass(eq=False)
class Inform7PredicatesNode(ModelBase):
    predicates: Any = None


@dataclass(eq=False)
class Inform7CommandNode(ModelBase):
    command: Any = None
    event: Any = None
    rule: Any = None


@dataclass(eq=False)
class Inform7CommandsNode(ModelBase):
    commands: Any = None


@dataclass(eq=False)
class Inform7CodeNode(ModelBase):
    code: Any = None


@dataclass(eq=False)
class Inform7Node(ModelBase):
    parts: Any = None


@dataclass(eq=False)
class TypeNode(ModelBase):
    name: Any = None
    parts: Any = None
    supertypes: Any = None


@dataclass(eq=False)
class DocumentNode(ModelBase):
    types: Any = None



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT license.


# -*- coding: utf-8 -*-
import sys
import textwrap
import subprocess

import importlib.resources
from typing import Union
from os.path import join as pjoin


from glk import ffi, lib
from io import StringIO

import textworld
from textworld.core import GameState
from textworld.core import GameNotRunningError

GLULX_PATH = pjoin(importlib.resources.files("textworld"), "thirdparty", "glulx", "Git-Glulx")


def _strip_input_prompt_symbol(text: str) -> str:
    if text.endswith("\n>"):
        return text[:-2]

    return text


class GitGlulxEnv(textworld.Environment):
    """ Environment to support playing Glulx games.

    This environment supports playing text-based games that were compiled for
    the `Glulx virtual machine <https://www.eblong.com/zarf/glulx>`_. The main
    advantage of using Glulx over Z-Machine is it uses 32-bit data and
    addresses, so it can handle game files up to four gigabytes long. This
    comes handy when we want to generate large world with a lot of objects
    in it.

    We use a customized version of `git-glulx <https://github.com/DavidKinder/Git>`_
    as the glulx interpreter. That way we don't rely on stdin/stdout to
    communicate with the interpreter but instead use UNIX sockets.

    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._process = None

    def close(self) -> None:
        if self.game_running:
            self._process.kill()
            self._process.wait()
            self._process = None

        try:
            lib.cleanup_glulx(self._names_struct)
        except AttributeError:
            pass  # Attempted to kill before reset

    def __del__(self):
        self.close()

    def load(self, ulx_file: str) -> None:
        # TODO check file format.
        self.close()  # Terminate existing process if needed.
        self._gamefile = ulx_file

    @property
    def game_running(self) -> bool:
        """ Determines if the game is still running. """
        return self._process is not None and self._process.poll() is None

    def step(self, command: str) -> str:
        if not self.game_running:
            raise GameNotRunningError()

        self.state = GameState()
        self.state.last_command = command.strip()
        self.state.raw = self._send(self.state.last_command)
        if self.state.raw is None:
            raise GameNotRunningError()

        self.state.feedback = _strip_input_prompt_symbol(self.state.raw)
        self.state.score = 0  # Default value.
        self.state.done = False  # Default value.
        return self.state, self.state.score, self.state.done

    def _send(self, command: str) -> Union[str, None]:
        """ Send a command directly to the interpreter.

        This method will not affect the internal state variable.
        """
        if not self.game_running:
            return None

        if len(command) == 0:
            command = " "

        c_command = ffi.new('char[]', command.encode('utf-8'))
        result = lib.communicate(self._names_struct, c_command)
        if result == ffi.NULL:
            self.close()
            return None

        result = ffi.gc(result, lib.free)
        return ffi.string(result).decode('utf-8')

    def reset(self) -> str:
        self.close()  # Terminate existing process if needed.

        self._names_struct = ffi.new('struct sock_names*')

        lib.init_glulx(self._names_struct)
        sock_name = ffi.string(self._names_struct.sock_name).decode('utf-8')
        self._process = subprocess.Popen(["%s/git-glulx-ml" % (GLULX_PATH,), self._gamefile, '-g', sock_name, '-q'])
        c_feedback = lib.get_output_nosend(self._names_struct)
        if c_feedback == ffi.NULL:
            self.close()
            raise ValueError("Game failed to start properly: {}.".format(self._gamefile))
        c_feedback = ffi.gc(c_feedback, lib.free)

        feedback = ffi.string(c_feedback).decode('utf-8')
        feedback = _strip_input_prompt_symbol(feedback)
        self.state = GameState(feedback=feedback, raw=feedback)
        return self.state

    def render(self, mode: str = "human") -> None:
        outfile = StringIO() if mode in ['ansi', "text"] else sys.stdout

        msg = self.state.feedback.rstrip() + "\n"
        if self.display_command_during_render and self.state.last_command is not None:
            msg = '> ' + self.state.last_command + "\n" + msg

        # Wrap each paragraph.
        if mode == "human":
            paragraphs = msg.split("\n")
            paragraphs = ["\n".join(textwrap.wrap(paragraph, width=80)) for paragraph in paragraphs]
            msg = "\n".join(paragraphs)

        outfile.write(msg + "\n")

        if mode == "text":
            outfile.seek(0)
            return outfile.read()

        if mode == 'ansi':
            return outfile



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT license.

import os
from os.path import join as pjoin
from subprocess import check_output

from textworld.utils import make_temp_directory

SCRIPTS_PATH = os.path.abspath(pjoin(__file__, "..", "..", "scripts"))


def test_check_generated_game():
    NB_GAMES = 3
    script = pjoin(SCRIPTS_PATH, "check_generated_games.py")
    with make_temp_directory(prefix="test_check_generated_game") as tmpdir:
        for i in range(NB_GAMES):
            command = ["tw-make", "custom", "--seed", str(i), "--output", tmpdir]
            check_output(command).decode()

        game_files = [pjoin(tmpdir, f) for f in os.listdir(tmpdir) if f.endswith(".z8")]
        assert len(game_files) == NB_GAMES
        command = ["python", script] + game_files
        stdout = check_output(command).decode()
        for file in game_files:
            assert "Testing " + file in stdout



#!/usr/bin/env python

# CAVEAT UTILITOR
#
# This file was automatically generated by TatSu.
#
#    https://pypi.python.org/pypi/tatsu/
#
# Any changes you make to it will be overwritten the next time
# the file is generated.

from __future__ import annotations

from typing import Any
from dataclasses import dataclass

from tatsu.objectmodel import Node
from tatsu.semantics import ModelBuilderSemantics


@dataclass(eq=False)
class ModelBase(Node):
    pass


class TextGrammarModelBuilderSemantics(ModelBuilderSemantics):
    def __init__(self, context=None, types=None):
        types = [
            t for t in globals().values()
            if type(t) is type and issubclass(t, ModelBase)
        ] + (types or [])
        super().__init__(context=context, types=types)


@dataclass(eq=False)
class Literal(ModelBase):
    value: Any = None


@dataclass(eq=False)
class AdjectiveNoun(ModelBase):
    adjective: Any = None
    noun: Any = None


@dataclass(eq=False)
class Match(ModelBase):
    lhs: Any = None
    rhs: Any = None


@dataclass(eq=False)
class ProductionRule(ModelBase):
    alternatives: Any = None
    symbol: Any = None


@dataclass(eq=False)
class TextGrammar(ModelBase):
    rules: Any = None



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT license.

from subprocess import check_call

import textworld
from textworld.utils import make_temp_directory


def test_playing_a_game():
    for ext in [".ulx", ".z8"]:
        with make_temp_directory(prefix="test_tw-play") as tmpdir:
            options = textworld.GameOptions()
            options.file_ext = ext
            options.path = tmpdir
            options.nb_rooms = 5
            options.nb_objects = 10
            options.quest_length = 5
            options.quest_breadth = 2
            options.seeds = 1234
            game_file, _ = textworld.make(options)

            command = ["tw-play", "--max-steps", "100", "--mode", "random", game_file]
            assert check_call(command) == 0

            command = ["tw-play", "--max-steps", "100", "--mode", "random-cmd", game_file]
            assert check_call(command) == 0

            command = ["tw-play", "--max-steps", "100", "--mode", "walkthrough", game_file]
            assert check_call(command) == 0



#!/usr/bin/env python

# CAVEAT UTILITOR
#
# This file was automatically generated by TatSu.
#
#    https://pypi.python.org/pypi/tatsu/
#
# Any changes you make to it will be overwritten the next time
# the file is generated.

from __future__ import annotations

from typing import Any
from dataclasses import dataclass

from tatsu.objectmodel import Node
from tatsu.semantics import ModelBuilderSemantics


@dataclass(eq=False)
class ModelBase(Node):
    pass


class PddlLogicModelBuilderSemantics(ModelBuilderSemantics):
    def __init__(self, context=None, types=None):
        types = [
            t for t in globals().values()
            if type(t) is type and issubclass(t, ModelBase)
        ] + (types or [])
        super().__init__(context=context, types=types)


@dataclass(eq=False)
class ActionTemplateNode(ModelBase):
    template: Any = None


@dataclass(eq=False)
class ActionFeedbackNode(ModelBase):
    name: Any = None


@dataclass(eq=False)
class ActionPddlNode(ModelBase):
    code: Any = None


@dataclass(eq=False)
class ActionGrammarNode(ModelBase):
    code: Any = None


@dataclass(eq=False)
class ActionTypeNode(ModelBase):
    feedback: Any = None
    grammar: Any = None
    name: Any = None
    pddl: Any = None
    template: Any = None


@dataclass(eq=False)
class PddlDocumentNode(ModelBase):
    parts: Any = None
